# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.

"""
Triton-based Multi-GPU Communication Tests

This package contains tests for Triton-based communication primitives using Iris.
"""
